#' Compute univariate statistics.
#'
#' \code{jk.means} computes moments- and rank-based non-robust and robust univariate statistics
#' and returns approximate results including a small confidentiality protection error, 
#' together with approximate information about this error.
#' 
#' Moments-based statistics reported:
#' 
#' \tabular{ll}{
#'      n       \tab no. of non-null values \cr
#'      sum     \tab sum \cr
#'      uss     \tab square sum \cr
#'      us3     \tab cube sum \cr
#'      us4     \tab sum of 4th powers \cr
#'      mean    \tab arithmetic mean \cr
#'      css     \tab corrected square sum \cr
#'      var     \tab variance \cr
#'      stddev  \tab standard deviation \cr
#'      stderrm \tab std. error of mean \cr
#'      t       \tab t test statistics \cr
#'      skew    \tab skewness \cr
#'      kurt    \tab kurtosis \cr
#'      bimod   \tab bimodality \cr
#'      cv      \tab coeff. of variation (pct.)\cr
#'      l4      \tab Blest's L4 location \cr
#'      bkurt   \tab Blest's kurtosis \cr
#' }
#' 
#' Quantiles reported:
#' 
#' \tabular{ll}{
#'      min2    \tab 2nd smallest \cr
#'      p1      \tab 1st percentile \cr
#'      p1_25   \tab 1.25th percentile \cr
#'      p2      \tab 2nd percentile \cr
#'      p2_5    \tab 2.5th percentile \cr
#'      p5      \tab 5th percentile \cr
#'      d1      \tab 1st decile \cr
#'      o1      \tab 1st octile \cr
#'      d2      \tab 2nd decile \cr
#'      q1      \tab lower quartile \cr
#'      d3      \tab 3rd decile \cr
#'      o3      \tab 3rd octile \cr
#'      d4      \tab 4th decile \cr
#'      median  \tab median \cr
#'      d6      \tab 6th decile \cr
#'      o5      \tab 5th octile \cr
#'      d7      \tab 7th decile \cr
#'      q3      \tab upper quartile \cr
#'      d8      \tab 8th decile \cr
#'      o7      \tab 7th octile \cr
#'      d9      \tab 9th decile \cr
#'      p95     \tab 95th percentile \cr
#'      p97_5   \tab 97.5th percentile \cr
#'      p98     \tab 98th percentile \cr
#'      p98_75  \tab 98.75th percentile \cr
#'      p99     \tab 99th percentile \cr
#'      max2    \tab 2nd largest \cr
#' }
#' 
#' Other derived statistics reported:
#' 
#' \tabular{ll}{
#'      trimean \tab Tukey's trimean \cr
#'      range   \tab range \cr
#'      qrange  \tab quartile range \cr
#'      lnotch  \tab lower boxplot notch \cr
#'      unotch  \tab upper boxplot notch \cr
#'      qmcoeff \tab quartile-median coeff. of var. \cr
#'      qcoeff  \tab pure quartile coeff. of var. \cr
#'      mskew   \tab Pearson's median-skewness \cr
#'      yskew   \tab Yule's skewness \cr
#'      qskew   \tab Bowley's quartile skewness \cr
#'      h10skew \tab Hinkley's 1st decile skewness \cr
#'      h5skew  \tab Hinkley's 5th percentile skewness \cr
#'      h1skew  \tab Hinkley's 1st percentile skewness \cr
#'      mkurt   \tab Moors' kurtosis \cr
#'      cskurt  \tab Crow-Siddiqui's kurtosis \cr
#' }
#' 
#' (All kurtosis variants are normalized to equal 0 for a normal distribution)
#' 
#' \code{jk.means(input.table=..., analyse=..., where=..., group.by=...)} 
#' reports the approximate value (v) and the approximate protection error (e) 
#' for the above univariate statistics,
#' for each variable specified in \code{analyse},
#' based on the set of observations in \code{input.table} that fulfil the optional \code{where} condition,
#' optionally grouped by the hierarchical grouping dimensions specified in \code{group.by}.
#' 
#' Confidence intervals for the true (sample) value of the statistics can be constructed from v and e:
#' [v- e,v+ e] is a 75\% confidence interval,
#' [v-2e,v+2e] is a 95\% confidence interval, and
#' [v-3e,v+3e] is a 99\% confidence interval.
#' 
#' @export
#' @param connection        MySQL database connection object as returned by \code{dbConnect}.
#' @param input.table       String naming the database table to analyse.
#' @param analyse           List specifying the analysis columns and the protection quartiles,
#'                          each entry being of the format \code{list(name="column name", quartiles=c(min,q1,median,q3,max))},
#'                          where \code{min,q1,median,q3,max} are real numbers that should resemble the 
#'                          true (population) quantiles of the variable
#'                          (\code{min} might be \code{-Inf}, \code{max} might be \code{Inf}).
#'                          If in doubt, estimate q1, median, and q3 from the sample, apply some rounding 
#'                          and use \code{quartiles=(-Inf, rounded sample q1, rounded sample median, rounded sample q3, Inf)} 
#' @param where             String containing a where condition (an SQL boolean expression used to filter rows).
#'                          Default: \code{"TRUE"}
#' @param group.by          List specifying the desired grouping in the format \code{list(dim,...,dim)},
#'                          each \code{dim} being of the format
#'                              \code{list(name="column name", label="...", groups=list(group, ..., group))}
#'                          and each \code{group} being of the format
#'                              \code{list(level=l, label="...", condition="sql boolean expression")},
#'                          where \code{l} is the no. of the hierarchical level, 1 being the outmost level,
#'                          and the where condition can contain the placeholder "\code{x}" in place of the column name.
#'                          Default: \code{list(list(name="all", label="all", groups=list(list(level=1, label="all", condition="TRUE"))))}
#' @param mu0               Comparison value for Student's one-sample t test.
#'                          Default: \code{0}
#' @param alpha.notch       Confidence level for boxplot notches.
#'                          Default: \code{.025}
#' @param return.format     Format of returned result: \code{"df"} for a linearized data frame, 
#'                          \code{"mudit"} for a nested list representing a multi-dimentional table (mudit format).
#'                          Default: \code{"df"}
#' @param debug             If \code{1}, a lot of debugging info is output. DON'T USE IN PRODUCTION SINCE IT MIGHT REVEAL CONFIDENTIAL DATA!
#'                          Default: \code{0}
#' @param secret            A string only known to the data provider and equal for all interrelated data tables of this provider.
#'                          CONFIDENTIALITY RELIES ON THIS BEING SECRET TO USERS!
#' @return Approximate results and approximate protection errors,
#'      either as a linearized data frame or as a nested list representing a multi-dimentional table (mudit format)
#' @callGraphPrimitives
#' @callGraphDepth 5
#' @references
#'      \url{http://www.unece.org/stats/documents/ece/ces/ge.46/2005/wp.39.e.pdf}
#'      \url{http://www.unece.org/stats/documents/ece/ces/ge.46/2005/wp.39.add.1.e.pdf}
#' @keywords univar robust
#' @author Jobst Heitzig \email{heitzig-j@@users.sourceforge.net}
#' @examples
#'      jk.means(connection=my.mysql.connection, input.table='person',
#'               analyse=list(list(name='age',label='Age in years',quartiles=c(0,20,40,60,Inf))))
#' 
jk.means <- function (
     connection
	,input.table
	,analyse
	,group.by      = list()
	,where	       = "TRUE"
	,mu0           = 0
    ,alpha.notch   = .025
    ,return.format = "df"
    ,debug         = 0
    ,secret        = ""
){
    if (secret=="") {
        cat ("ERROR: no secret given! Please specify a string only known to the data provider and equal for all interrelated data tables of this provider!\n")
        return (NULL)
    }
    
	# prepare everything:
	
    .g <- list(
         mysql.connection = connection
        ,.jk.debug = debug
        ,.jk.secret = secret
    	,input.table = input.table
    	,analyse = analyse
    	,group.by = group.by
    	,where = where
    )	
	.g <- .jk.univariate.prepare(.g)


	# handy abbreviations:
	
	x <- .g$analyse.names
	x.sql <- cpaste(x)
	g <- .g$group.by.ids
	g.sql <- .g$group.by.ids.sql
	g.t.sql <- cpaste("t.",g)

    notch.factor <- 1.58/qnorm(.975)*qnorm(1-alpha.notch)
    
    
    # 1. MOMENT-BASED STATS:

    
	# prepare db:
	
	dbSendQuery(.g$mysql.connection, "DROP TABLE IF EXISTS _powersums" )
	dbSendQuery(.g$mysql.connection, "DROP TABLE IF EXISTS _trueresults1" )


	# let db compute true power sums:
	
	dbSendQuery(.g$mysql.connection,cpaste(
	
		 " CREATE TABLE _powersums "
		," AS SELECT ",g.sql,", "
			,cpaste(
				 "  COUNT(",  x,")    AS n__",  x  # no. of non-null values
				," ,SUM(",    x,")    AS sum__",x
				," ,SUM(POW(",x,",2)) AS uss__",x  # uncorrected sum of squares
				," ,SUM(POW(",x,",3)) AS us3__",x  # ... of cubes
				," ,SUM(POW(",x,",4)) AS us4__",x  # ... of 4th powers
			)
		," FROM ",.g$grouped.table," AS t "
		," WHERE ",.g$where
		," GROUP BY ",g.sql
	))

	if (.g$.jk.debug) {
		print("_powersums")
		print(dbReadTable(.g$mysql.connection,"_powersums"))
	}
	

	# let db compute true stats:
	
	stat.names1 <- 
		c(
			 "n","sum","uss","us3","us4"
			,"mean","css","var","stddev","stderrm"
			,"t","skew","kurt","bimod","cv"
            ,"l4","bkurt"
		)
		
	stat.labels1 <- 
		c(
			 "no. of non-null values","sum","square sum","cube sum","sum of 4th powers"
			,"arithmetic mean","corrected square sum","variance","standard deviation","std. error of mean"
			,"t test statistics","skewness","kurtosis","bimodality","coeff. of variation (pct.)"
            ,"Blest's L4 location","Blest's kurtosis"
		)
		
	stat.formulas1 <- gsub("[ ]+"," ",cpaste(
#################################################

# mean = sum / n :

	 "  sum__",x,"/n__",x," AS mean__",x

# css = uss - sum²/n :

	," ,uss__",x,"-POW(sum__",x,",2)/n__",x," AS css__",x

# var[iance] = css / (n-1) 
#            = (uss - sum²/n) / (n-1) :

	," ,(uss__",x,"-POW(sum__",x,",2)/n__",x,")/(n__",x,"-1) AS var__",x

# stddev = sqrt(var) 
#        = sqrt( (uss - sum²/n) / (n-1) ) :

	," ,SQRT((uss__",x,"-POW(sum__",x,",2)/n__",x,")/(n__",x,"-1)) AS stddev__",x

# stderrm = stddev / sqrt(n) 
#         = sqrt( (uss - sum²/n) / (n-1)n ) :

	," ,SQRT((uss__",x,"-POW(sum__",x,",2)/n__",x,")/(n__",x,"-1)/n__",x,") AS stderrm__",x

# t = (mean - mu0) / stderr
#   = (sum/n - mu0) / sqrt( (uss - sum²/n) / (n-1)n ) :

	," ,(sum__",x,"/n__",x,"-",mu0,")/SQRT((uss__",x,"-POW(sum__",x,",2)/n__",x,")/(n__",x,"-1)/n__",x,") AS t__",x

# skew[ness] = (us3*n + 2sum³/n - 3uss*sum) / stddev(n-2)css
#            = (us3*n + 2sum³/n - 3uss*sum) / sqrt( (uss - sum²/n) / (n-1) )(n-2)(uss - sum²/n) :

	," ,(us3__",x,"*n__",x,"+2*POW(sum__",x,",3)/n__",x,"-3*uss__",x,"*sum__",x,")/SQRT((uss__",x,"-POW(sum__",x,",2)/n__",x,")/(n__",x,"-1))/(n__",x,"-2)/(uss__",x,"-POW(sum__",x,",2)/n__",x,") AS skew__",x

# kurt[osis] = ( (n+1)n(n-1)(us4 + mean ( 3mean(uss+css) - 4us3 )) / css² - 3(n-1)²) / (n-2)(n-3)
#            = ( (n+1)n(n-1)(us4 + sum/n ( sum/n (6uss-3sum²/n) - 4us3 )) / (uss - sum²/n)² - 3(n-1)²) / (n-2)(n-3) :

	," ,((n__",x,"+1)*n__",x,"*(n__",x,"-1)*(us4__",x,"+sum__",x,"/n__",x,"*(sum__",x,"/n__",x,"*(6*uss__",x,"-3*POW(sum__",x,",2)/n__",x,")-4*us3__",x,"))/POW(uss__",x,"-POW(sum__",x,",2)/n__",x,",2)-3*POW(n__",x,"-1,2))/(n__",x,"-2)/(n__",x,"-3) AS kurt__",x
		
# bimod[ality] = ... :

	," ,((us3__",x,"*n__",x,"+2*POW(sum__",x,",3)/n__",x,"-3*uss__",x,"*sum__",x,")/SQRT((uss__",x,"-POW(sum__",x,",2)/n__",x,")/(n__",x,"-1))/(n__",x,"-2)/(uss__",x,"-POW(sum__",x,",2)/n__",x,"))/(((n__",x,"+1)*n__",x,"*(n__",x,"-1)*(us4__",x,"+sum__",x,"/n__",x,"*(sum__",x,"/n__",x,"*(6*uss__",x,"-3*POW(sum__",x,",2)/n__",x,")-4*us3__",x,"))/POW(uss__",x,"-POW(sum__",x,",2)/n__",x,",2)-3*POW(n__",x,"-1,2))/(n__",x,"-2)/(n__",x,"-3)+3*POW(n__",x,"-1,2)/(n__",x,"-2)/(n__",x,"-3))"," AS bimod__",x

# cv = ... :

	," ,100*sqrt((uss__",x,"-POW(sum__",x,",2)/n__",x,")/(n__",x,"-1))/(sum__",x,"/n__",x,")"," AS cv__",x

# l4 = mean + stddev*blestf  with  blestf = (sqrt(skew²/4+1)+skew/2)^(1/3) - (sqrt(skew²/4+1)-skew/2)^(1/3) :

    ," ,sum__",x,"/n__",x,"+SQRT((uss__",x,"-POW(sum__",x,",2)/n__",x,")/(n__",x,"-1))*"
        ,"( POW(SQRT(POW(((us3__",x,"*n__",x,"+2*POW(sum__",x,",3)/n__",x,"-3*uss__",x,"*sum__",x,")/SQRT((uss__",x,"-POW(sum__",x,",2)/n__",x,")/(n__",x,"-1))/(n__",x,"-2)/(uss__",x,"-POW(sum__",x,",2)/n__",x,")),2)/4+1)+((us3__",x,"*n__",x,"+2*POW(sum__",x,",3)/n__",x,"-3*uss__",x,"*sum__",x,")/SQRT((uss__",x,"-POW(sum__",x,",2)/n__",x,")/(n__",x,"-1))/(n__",x,"-2)/(uss__",x,"-POW(sum__",x,",2)/n__",x,"))/2,1/3)-POW(SQRT(POW(((us3__",x,"*n__",x,"+2*POW(sum__",x,",3)/n__",x,"-3*uss__",x,"*sum__",x,")/SQRT((uss__",x,"-POW(sum__",x,",2)/n__",x,")/(n__",x,"-1))/(n__",x,"-2)/(uss__",x,"-POW(sum__",x,",2)/n__",x,")),2)/4+1)-((us3__",x,"*n__",x,"+2*POW(sum__",x,",3)/n__",x,"-3*uss__",x,"*sum__",x,")/SQRT((uss__",x,"-POW(sum__",x,",2)/n__",x,")/(n__",x,"-1))/(n__",x,"-2)/(uss__",x,"-POW(sum__",x,",2)/n__",x,"))/2,1/3) )"
        ," AS l4__",x

# bkurt = kurt - 3((1+blestf²)²-1) :

    ," ,((n__",x,"+1)*n__",x,"*(n__",x,"-1)*(us4__",x,"+sum__",x,"/n__",x,"*(sum__",x,"/n__",x,"*(6*uss__",x,"-3*POW(sum__",x,",2)/n__",x,")-4*us3__",x,"))/POW(uss__",x,"-POW(sum__",x,",2)/n__",x,",2)-3*POW(n__",x,"-1,2))/(n__",x,"-2)/(n__",x,"-3)"
        ,"-3*(POW(1+POW(" 
        ,"( POW(SQRT(POW(((us3__",x,"*n__",x,"+2*POW(sum__",x,",3)/n__",x,"-3*uss__",x,"*sum__",x,")/SQRT((uss__",x,"-POW(sum__",x,",2)/n__",x,")/(n__",x,"-1))/(n__",x,"-2)/(uss__",x,"-POW(sum__",x,",2)/n__",x,")),2)/4+1)+((us3__",x,"*n__",x,"+2*POW(sum__",x,",3)/n__",x,"-3*uss__",x,"*sum__",x,")/SQRT((uss__",x,"-POW(sum__",x,",2)/n__",x,")/(n__",x,"-1))/(n__",x,"-2)/(uss__",x,"-POW(sum__",x,",2)/n__",x,"))/2,1/3)-POW(SQRT(POW(((us3__",x,"*n__",x,"+2*POW(sum__",x,",3)/n__",x,"-3*uss__",x,"*sum__",x,")/SQRT((uss__",x,"-POW(sum__",x,",2)/n__",x,")/(n__",x,"-1))/(n__",x,"-2)/(uss__",x,"-POW(sum__",x,",2)/n__",x,")),2)/4+1)-((us3__",x,"*n__",x,"+2*POW(sum__",x,",3)/n__",x,"-3*uss__",x,"*sum__",x,")/SQRT((uss__",x,"-POW(sum__",x,",2)/n__",x,")/(n__",x,"-1))/(n__",x,"-2)/(uss__",x,"-POW(sum__",x,",2)/n__",x,"))/2,1/3) )"
        ,",2),2)-1) AS bkurt__",x

#####################################
	))
	
	dbSendQuery(.g$mysql.connection,cpaste(
	
		 " CREATE TABLE _trueresults1 "
		," AS SELECT *, ",stat.formulas1," FROM _powersums "
	))

	if (.g$.jk.debug) {
		print("_trueresults1")
		print(dbReadTable(.g$mysql.connection,"_trueresults1"))
	}
	

	# let db compute changes of all needed statistics
	# when removing one value:
	# (variants in which one value is replaced lead to smaller deviations and
	# can thus be skipped)
	
	powersum.variants.view <- cpaste(
	
		 " ( SELECT ",g.t.sql,", "
			,cpaste(
				 "  n__",  x,           "-1 AS n__",  x
				," ,sum__",x,"-",    x,   " AS sum__",x
				," ,uss__",x,"-POW(",x,",2) AS uss__",x
				," ,us3__",x,"-POW(",x,",3) AS us3__",x
				," ,us4__",x,"-POW(",x,",4) AS us4__",x
			)
		," FROM _powersums AS p, ",.g$grouped.table," AS t"
		," WHERE ",.g$where," AND "
			,paste(
				 "p.",g,"=t.",g
				,sep="",collapse=" AND ")
		," ) "
	)
	
	if (.g$.jk.debug) {
		print("powersum.variants.view")
		print(powersum.variants.view)
		print(dbGetQuery(.g$mysql.connection,powersum.variants.view))
	}
	
	.g$result.variants.view <- 
		cpaste(
	
		" ( SELECT *, ",stat.formulas1," FROM ",powersum.variants.view," AS v ) "
	)
	
	if (.g$.jk.debug) {
		print("result.variants.view")
		print(.g$result.variants.view)
		print(dbGetQuery(.g$mysql.connection,.g$result.variants.view))
	}


    # let db compute min and max variants:
    
    s0 <- rep(stat.names1,rep(length(x),length(stat.names1)))
    x0 <- rep(x,length(stat.names1))
    
    dbSendQuery(.g$mysql.connection, "DROP TABLE IF EXISTS _stats1_extreme_variants" )
    dbSendQuery(.g$mysql.connection,cpaste(
                    
         " CREATE TABLE _stats1_extreme_variants "
        ," AS SELECT ",g.sql,", "
            ,cpaste(
                 "  MIN(",s0,"__",x0,") AS l_",s0,"__",x0
                ,", MAX(",s0,"__",x0,") AS u_",s0,"__",x0
            )
        ," FROM ",.g$result.variants.view," AS t"
        ," GROUP BY ",g.sql
    ))
    stats1.extreme.variants.sql <- cpaste(unlist(lapply(stat.names1,function(si){lapply(c("l_","u_"),function(vi){lapply(x,function(xi){cpaste(vi,si,"__",xi)})})})))
        
    
    # 2. QUANTILES:


    # let db compute data required for quantiles:
    
    for (i in 1:length(x)) {
        
        xi <- x[[i]]
        
       
        # let db sort values in groups:
    
        dbSendQuery(.g$mysql.connection, "DROP TABLE IF EXISTS _ordered" )
        dbSendQuery(.g$mysql.connection,cpaste(
                        
             " CREATE TABLE _ordered (id__ INT AUTO_INCREMENT PRIMARY KEY) "
            ," AS SELECT ",g.sql,", ",xi
            ," FROM ",.g$grouped.table," AS t "
            ," WHERE (",.g$where,") AND (",xi," IS NOT NULL) "
            ," ORDER BY ",g.sql,", ",xi
        ))
    
        # get no. non-null values and group starting row-ids in table:
    
        ids.df <- dbGetQuery(.g$mysql.connection,cpaste(
            
            " SELECT ",g.sql
            ,", MIN(id__) AS mi"
            ,", COUNT(*) AS n"
            ," FROM _ordered AS t "
            ," GROUP BY ",g.sql
            ," HAVING n >= 4 "
        ))
        
        # compose list of row-ids to fetch:

        stat.names2 <- 
            c(
                 "min2","p1","p1_25","p2","p2_5"
                ,"p5","d1","o1","d2"
                ,"q1","d3","o3","d4"
                ,"median","d6","o5","d7"
                ,"q3","d8","o7","d9"
                ,"p95","p97_5","p98","p98_75","p99"
                ,"max2"
            )
            
        stat.labels2 <- 
            c(
                 "2nd smallest","1st percentile","1.25th percentile","2nd percentile","2.5th percentile"
                ,"5th percentile","1st decile","1st octile","2nd decile"
                ,"lower quartile","3rd decile","3rd octile","4th decile"
                ,"median","6th decile","5th octile","7th decile"
                ,"upper quartile","8th decile","7th octile","9th decile"
                ,"95th percentile","97.5th percentile","98th percentile","98.75th percentile","99th percentile"
                ,"2nd largest"
            )

        q <- c(	 0,.01,.0125,.02,.025
                ,.05,.1,.125,.2
                ,.25,.3,.375,.4
                ,.5,.6,.625,.7
                ,.75,.8,.875,.9
                ,.95,.975,.98,.9875,.99
                ,1)

        quantiles.sql <- cpaste(unlist(lapply(stat.names2,function(si){lapply(c("l_","","u_"),function(vi){lapply(x,function(xi){cpaste(vi,si,"__",xi)})})})))
        
        nq <- length(q)
        ng <- length(ids.df$n)
        repq <- rep(q,ng)
        repnqng <- rep(nq,ng)
        repn1 <- rep(unlist(ids.df$n)+1,repnqng)
        repn1repq <- repn1*repq
        f0 <- floor(repn1repq)
        f <- ifelse(2>f0,2,ifelse(repn1-3<f0,repn1-3,f0)) # rel. pos. of lower value for quantile estimation
        remainder <- ifelse(2>f0,0,ifelse(repn1-3<f0,1,repn1repq-f0)) # remainder used for weighting
        u <- c(rep(unlist(ids.df$mi),repnqng)+f) # pos. of upper value for quantile estimation
        nu <- length(u)
        ids.to.fetch <- rep(u,rep(4,nu)) + rep(c(-2,-1,0,1),nu) # ids of all needed values
    
        # fetch those values from db:

        dbWriteTable(.g$mysql.connection,"_ids_to_fetch"
            ,data.frame(list(
                 rownum__ = 1:length(ids.to.fetch)
                ,id__ = ids.to.fetch
                ))
            ,overwrite=TRUE)
        
        rm(f0,f,u,ids.to.fetch)
        
        values.df <- dbGetQuery(.g$mysql.connection,cpaste(
            
             " SELECT rownum__, ",xi," AS x "
            ," FROM _ordered AS o, _ids_to_fetch AS f "
            ," WHERE o.id__ = f.id__ "
            ," ORDER BY rownum__ "
        ))
        
        # compute quantile estimators and min and max variants:

        off <- 4*(0:(nu-1))
        low.nb  <- values.df$x[off+1]
        low.val <- values.df$x[off+2]
        up.val  <- values.df$x[off+3]
        up.nb   <- values.df$x[off+4]
        
        long.list <- list()
        for (gi in .g$group.by.ids) long.list[[gi]] <- rep(ids.df[[gi]],repnqng)
        long.list[["stat__"]] <- rep(stat.names2,ng)
        long.df <- data.frame(long.list)
        long.df <- data.frame(long.list)
        long.df[[cpaste("l__",xi)]] <- remainder*low.val + (1-remainder)*low.nb
        long.df[[cpaste("__",xi)]]  <- remainder*up.val + (1-remainder)*low.val
        long.df[[cpaste("u__",xi)]] <- remainder*up.nb + (1-remainder)*up.val

        rm(remainder,low.val,low.nb,up.val,up.nb,ids.df,long.list)

        wide.df <- reshape(long.df,direction="wide",idvar=.g$group.by.ids,timevar="stat__")
        for (sni in stat.names2) {
            colnames(wide.df)[colnames(wide.df)==cpaste("l__",xi,".",sni)] <- cpaste("l_",sni,"__",xi)
            colnames(wide.df)[colnames(wide.df)==cpaste("__",xi,".",sni)]  <- cpaste(sni,"__",xi)
            colnames(wide.df)[colnames(wide.df)==cpaste("u__",xi,".",sni)] <- cpaste("u_",sni,"__",xi)
        }

        # possible extension: optionally produce plots, e.g. boxplots:
        if (0) {
            bxp.z <- list(
                         stats=t(as.matrix(wide.df[c(
                             cpaste("min2__",xi)
                            ,cpaste("q1__",xi)
                            ,cpaste("median__",xi)
                            ,cpaste("q3__",xi)
                            ,cpaste("max2__",xi)
                        )]))
                        ,n=rep(101,ng)
                        ,conf=t(as.matrix(wide.df[c(
                             cpaste("q1__",xi)
                            ,cpaste("q3__",xi)
                        )]))
                        ,out=rep(0,ng) #?
                        ,group=rep(0,ng) #?
                        ,names=rep("",ng) #?
                        )
            bxp(bxp.z,outline=0)
        }
        
        # merge this variable's data into the overall quantiles df:

        if (i==1) quantiles.and.variants.df <- wide.df
        else quantiles.and.variants.df <- merge(quantiles.and.variants.df, wide.df, by=.g$group.by.ids, all.x=TRUE, all.y=TRUE)
        
        # free unused mem:

        rm(wide.df)
        gc()
    }

    # upload quantiles df to db:

    dbWriteTable(.g$mysql.connection,"_quantiles_and_variants"
        ,quantiles.and.variants.df
        ,overwrite=TRUE)
    rm(quantiles.and.variants.df)
    gc()

    
    # 3. DERIVED STATISTICS:


    # merge moment-based and rank-base tables and
    # compute true values and extreme variants of derives stats:

    stat.names3 <- 
        c(
             "trimean" # Future work: "biweight" (uses MAD)
            ,"range","qrange","lnotch","unotch" # Future work: MAD
            ,"qmcoeff","qcoeff"
            ,"mskew","yskew","qskew"
            ,"h10skew","h5skew","h1skew" 
            ,"mkurt","cskurt"
        )
        
    stat.labels3 <- 
        c(
             "Tukey's trimean" # ,"Tukey's biweight"
            ,"range","quartile range","lower boxplot notch","upper boxplot notch" # ,"median abs. dev. from median"
            ,"quartile-median coeff. of var.","pure quartile coeff. of var."
            ,"Pearson's median-skewness","Yule's skewness","Bowley's quartile skewness"
            ,"Hinkley's 1st decile skewness","Hinkley's 5th percentile skewness","Hinkley's 1st percentile skewness"
            ,"Moors' kurtosis","Crow-Siddiqui's kurtosis"
        )
        
    stat.formulas3 <- gsub("[ ]+"," ",cpaste(
#################################################

# trimean = (q1+2median+q3)/4 :
    
     "  (l_q1__",x,"+2*l_median__",x,"+l_q3__",x,")/4 AS l_trimean__",x
    ," ,(  q1__",x,"+2*  median__",x,"+  q3__",x,")/4 AS   trimean__",x
    ," ,(u_q1__",x,"+2*u_median__",x,"+u_q3__",x,")/4 AS u_trimean__",x

# range = max2-min2 :

    ," ,LEAST   (l_max2__",x,"-min2__",x,",max2__",x,"-u_min2__",x,") AS l_range__",x
    ," ,           max2__",x,"-min2__",x,"                            AS   range__",x
    ," ,GREATEST(u_max2__",x,"-min2__",x,",max2__",x,"-l_min2__",x,") AS u_range__",x

# qrange = q3-q1 :

    ," ,LEAST   (l_q3__",x,"-q1__",x,",q3__",x,"-u_q1__",x,") AS l_qrange__",x
    ," ,           q3__",x,"-q1__",x,"                        AS   qrange__",x
    ," ,GREATEST(u_q3__",x,"-q1__",x,",q3__",x,"-l_q1__",x,") AS u_qrange__",x

# lnotch = median - notchfactor*qrange/sqrt(n) :

    ," ,LEAST   (median__",x,"-",notch.factor,"*(u_q3__",x,"-  q1__",x,")/SQRT(l_n__",x,"),"
             ,"l_median__",x,"-",notch.factor,"*(  q3__",x,"-l_q1__",x,")/SQRT(l_n__",x,")) AS l_lnotch__",x
    ," ,         median__",x,"-",notch.factor,"*(  q3__",x,"-  q1__",x,")/SQRT(  n__",x,")  AS   lnotch__",x
    ," ,GREATEST(median__",x,"-",notch.factor,"*(l_q3__",x,"-  q1__",x,")/SQRT(u_n__",x,"),"
             ,"u_median__",x,"-",notch.factor,"*(  q3__",x,"-u_q1__",x,")/SQRT(u_n__",x,")) AS u_lnotch__",x

# unotch = median + notchfactor*qrange/sqrt(n) :

    ," ,LEAST   (median__",x,"+",notch.factor,"*(  q3__",x,"-u_q1__",x,")/SQRT(u_n__",x,"),"
             ,"l_median__",x,"+",notch.factor,"*(l_q3__",x,"-  q1__",x,")/SQRT(u_n__",x,")) AS l_unotch__",x
    ," ,         median__",x,"+",notch.factor,"*(  q3__",x,"-  q1__",x,")/SQRT(  n__",x,")  AS   unotch__",x
    ," ,GREATEST(median__",x,"+",notch.factor,"*(  q3__",x,"-l_q1__",x,")/SQRT(l_n__",x,"),"
             ,"u_median__",x,"+",notch.factor,"*(u_q3__",x,"-  q1__",x,")/SQRT(l_n__",x,")) AS u_unotch__",x

# qmcoeff = qrange/|median| :
        
    ," ,LEAST   ((l_q3__",x,"-  q1__",x,")/GREATEST(ABS(median__",x,"),ABS(l_median__",x,")),"
               ,"(  q3__",x,"-u_q1__",x,")/GREATEST(ABS(median__",x,"),ABS(u_median__",x,"))) AS l_qmcoeff__",x
    ," ,         (  q3__",x,"-  q1__",x,")/ABS(median__",x,")                                 AS   qmcoeff__",x
    ," ,GREATEST((  q3__",x,"-l_q1__",x,")/LEAST   (ABS(median__",x,"),ABS(l_median__",x,")),"
               ,"(u_q3__",x,"-  q1__",x,")/LEAST   (ABS(median__",x,"),ABS(u_median__",x,"))) AS u_qmcoeff__",x

# qcoeff = 2qrange/|q3+q1| :
        
    ," ,2*LEAST   ("
               ," (u_q3__",x,"-  q1__",x,")/ABS(u_q3__",x,"+  q1__",x,")"
               ,",(l_q3__",x,"-  q1__",x,")/ABS(l_q3__",x,"+  q1__",x,")"
               ,",(  q3__",x,"-u_q1__",x,")/ABS(  q3__",x,"+u_q1__",x,")"
               ,",(  q3__",x,"-l_q1__",x,")/ABS(  q3__",x,"+l_q1__",x,")"
               ,",(u_q3__",x,"-u_q1__",x,")/ABS(u_q3__",x,"+u_q1__",x,")"
               ,",(l_q3__",x,"-l_q1__",x,")/ABS(l_q3__",x,"+l_q1__",x,")"
                                                                     ,") AS l_qcoeff__",x
    ," ,2*        (  q3__",x,"-  q1__",x,")/ABS(  q3__",x,"+  q1__",x,") AS   qcoeff__",x
    ," ,2*GREATEST("
               ," (u_q3__",x,"-  q1__",x,")/ABS(u_q3__",x,"+  q1__",x,")"
               ,",(l_q3__",x,"-  q1__",x,")/ABS(l_q3__",x,"+  q1__",x,")"
               ,",(  q3__",x,"-u_q1__",x,")/ABS(  q3__",x,"+u_q1__",x,")"
               ,",(  q3__",x,"-l_q1__",x,")/ABS(  q3__",x,"+l_q1__",x,")"
               ,",(u_q3__",x,"-u_q1__",x,")/ABS(u_q3__",x,"+u_q1__",x,")"
               ,",(l_q3__",x,"-l_q1__",x,")/ABS(l_q3__",x,"+l_q1__",x,")"
                                                                     ,") AS u_qcoeff__",x

# mskew = 3(mean-median)/stddev :

    ," ,3*LEAST   ((l_mean__",x,"-u_median__",x,")/u_stddev__",x,","
                 ,"(l_mean__",x,"-u_median__",x,")/l_stddev__",x,") AS l_mskew__",x
    ," ,3*         (  mean__",x,"-  median__",x,")/  stddev__",x,"  AS   mskew__",x
    ," ,3*GREATEST((u_mean__",x,"-l_median__",x,")/u_stddev__",x,","
                 ,"(u_mean__",x,"-l_median__",x,")/l_stddev__",x,") AS u_mskew__",x

# yskew = (q1 - 2mean + q3)/qrange :

    ," ,LEAST   ("
               ," (l_q1__",x,"-2*  mean__",x,"+l_q3__",x,")/(l_q3__",x,"-  q1__",x,")"
               ,",(  q1__",x,"-2*u_mean__",x,"+  q3__",x,")/(u_q3__",x,"-  q1__",x,")"
               ,",(l_q1__",x,"-2*  mean__",x,"+l_q3__",x,")/(  q3__",x,"-l_q1__",x,")"
               ,",(  q1__",x,"-2*u_mean__",x,"+  q3__",x,")/(  q3__",x,"-u_q1__",x,")"
                                                                                  ,") AS l_yskew__",x
    ," ,          (  q1__",x,"-2*  mean__",x,"+  q3__",x,")/(  q3__",x,"-  q1__",x,") AS   yskew__",x
    ," ,GREATEST("
               ," (u_q1__",x,"-2*  mean__",x,"+u_q3__",x,")/(u_q3__",x,"-  q1__",x,")"
               ,",(  q1__",x,"-2*l_mean__",x,"+  q3__",x,")/(l_q3__",x,"-  q1__",x,")"
               ,",(u_q1__",x,"-2*  mean__",x,"+u_q3__",x,")/(  q3__",x,"-u_q1__",x,")"
               ,",(  q1__",x,"-2*l_mean__",x,"+  q3__",x,")/(  q3__",x,"-l_q1__",x,")"
                                                                                  ,") AS u_yskew__",x

# qskew = (q1 - 2median + q3)/qrange :

    ," ,LEAST   ("
               ," (l_q1__",x,"-2*  median__",x,"+l_q3__",x,")/(l_q3__",x,"-  q1__",x,")"
               ,",(  q1__",x,"-2*u_median__",x,"+  q3__",x,")/(u_q3__",x,"-  q1__",x,")"
               ,",(l_q1__",x,"-2*  median__",x,"+l_q3__",x,")/(  q3__",x,"-l_q1__",x,")"
               ,",(  q1__",x,"-2*u_median__",x,"+  q3__",x,")/(  q3__",x,"-u_q1__",x,")"
                                                                                    ,") AS l_qskew__",x
    ," ,          (  q1__",x,"-2*  median__",x,"+  q3__",x,")/(  q3__",x,"-  q1__",x,") AS   qskew__",x
    ," ,GREATEST("
               ," (u_q1__",x,"-2*  median__",x,"+u_q3__",x,")/(u_q3__",x,"-  q1__",x,")"
               ,",(  q1__",x,"-2*l_median__",x,"+  q3__",x,")/(l_q3__",x,"-  q1__",x,")"
               ,",(u_q1__",x,"-2*  median__",x,"+u_q3__",x,")/(  q3__",x,"-u_q1__",x,")"
               ,",(  q1__",x,"-2*l_median__",x,"+  q3__",x,")/(  q3__",x,"-l_q1__",x,")"
                                                                                    ,") AS u_qskew__",x

# h10skew = (d1 - 2median + d9)/(d9-d1) :

    ," ,LEAST   ("
               ," (l_d1__",x,"-2*  median__",x,"+l_d9__",x,")/(l_d9__",x,"-  d1__",x,")"
               ,",(  d1__",x,"-2*u_median__",x,"+  d9__",x,")/(u_d9__",x,"-  d1__",x,")"
               ,",(l_d1__",x,"-2*  median__",x,"+l_d9__",x,")/(  d9__",x,"-l_d1__",x,")"
               ,",(  d1__",x,"-2*u_median__",x,"+  d9__",x,")/(  d9__",x,"-u_d1__",x,")"
                                                                                    ,") AS l_h10skew__",x
    ," ,          (  d1__",x,"-2*  median__",x,"+  d9__",x,")/(  d9__",x,"-  d1__",x,") AS   h10skew__",x
    ," ,GREATEST("
               ," (u_d1__",x,"-2*  median__",x,"+u_d9__",x,")/(u_d9__",x,"-  d1__",x,")"
               ,",(  d1__",x,"-2*l_median__",x,"+  d9__",x,")/(l_d9__",x,"-  d1__",x,")"
               ,",(u_d1__",x,"-2*  median__",x,"+u_d9__",x,")/(  d9__",x,"-u_d1__",x,")"
               ,",(  d1__",x,"-2*l_median__",x,"+  d9__",x,")/(  d9__",x,"-l_d1__",x,")"
                                                                                    ,") AS u_h10skew__",x

# h5skew = (p5 - 2median + p95)/(p95-p5) :

    ," ,LEAST   ("
               ," (l_p5__",x,"-2*  median__",x,"+l_p95__",x,")/(l_p95__",x,"-  p5__",x,")"
               ,",(  p5__",x,"-2*u_median__",x,"+  p95__",x,")/(u_p95__",x,"-  p5__",x,")"
               ,",(l_p5__",x,"-2*  median__",x,"+l_p95__",x,")/(  p95__",x,"-l_p5__",x,")"
               ,",(  p5__",x,"-2*u_median__",x,"+  p95__",x,")/(  p95__",x,"-u_p5__",x,")"
                                                                                      ,") AS l_h5skew__",x
    ," ,          (  p5__",x,"-2*  median__",x,"+  p95__",x,")/(  p95__",x,"-  p5__",x,") AS   h5skew__",x
    ," ,GREATEST("
               ," (u_p5__",x,"-2*  median__",x,"+u_p95__",x,")/(u_p95__",x,"-  p5__",x,")"
               ,",(  p5__",x,"-2*l_median__",x,"+  p95__",x,")/(l_p95__",x,"-  p5__",x,")"
               ,",(u_p5__",x,"-2*  median__",x,"+u_p95__",x,")/(  p95__",x,"-u_p5__",x,")"
               ,",(  p5__",x,"-2*l_median__",x,"+  p95__",x,")/(  p95__",x,"-l_p5__",x,")"
                                                                                      ,") AS u_h5skew__",x

# h1skew = (p1 - 2median + p99)/(p99-p1) :

    ," ,LEAST   ("
               ," (l_p1__",x,"-2*  median__",x,"+l_p99__",x,")/(l_p99__",x,"-  p1__",x,")"
               ,",(  p1__",x,"-2*u_median__",x,"+  p99__",x,")/(u_p99__",x,"-  p1__",x,")"
               ,",(l_p1__",x,"-2*  median__",x,"+l_p99__",x,")/(  p99__",x,"-l_p1__",x,")"
               ,",(  p1__",x,"-2*u_median__",x,"+  p99__",x,")/(  p99__",x,"-u_p1__",x,")"
                                                                                      ,") AS l_h1skew__",x
    ," ,          (  p1__",x,"-2*  median__",x,"+  p99__",x,")/(  p99__",x,"-  p1__",x,") AS   h1skew__",x
    ," ,GREATEST("
               ," (u_p1__",x,"-2*  median__",x,"+u_p99__",x,")/(u_p99__",x,"-  p1__",x,")"
               ,",(  p1__",x,"-2*l_median__",x,"+  p99__",x,")/(l_p99__",x,"-  p1__",x,")"
               ,",(u_p1__",x,"-2*  median__",x,"+u_p99__",x,")/(  p99__",x,"-u_p1__",x,")"
               ,",(  p1__",x,"-2*l_median__",x,"+  p99__",x,")/(  p99__",x,"-l_p1__",x,")"
                                                                                      ,") AS u_h1skew__",x

# mkurt = (o7 - o5 + o3 - o1)/qrange - 1.23309511548521 : (see Moors, 1988)

    ," ,LEAST   ("
               ," (l_o7__",x,"-  o5__",x,"+  o3__",x,"-  o1__",x,")/(  q3__",x,"-  q1__",x,")"
               ,",(l_o7__",x,"-l_o5__",x,"+l_o3__",x,"-  o1__",x,")/(l_q3__",x,"-l_q1__",x,")"
               ,",(  o7__",x,"-  o5__",x,"+l_o3__",x,"-  o1__",x,")/(  q3__",x,"-l_q1__",x,")"
               ,",(  o7__",x,"-  o5__",x,"+  o3__",x,"-u_o1__",x,")/(  q3__",x,"-  q1__",x,")"
               ,",(  o7__",x,"-u_o5__",x,"+u_o3__",x,"-u_o1__",x,")/(u_q3__",x,"-u_q1__",x,")"
               ,",(  o7__",x,"-u_o5__",x,"+  o3__",x,"-  o1__",x,")/(u_q3__",x,"-  q1__",x,")"
                                                                                          ,")-1.23309511548521 AS l_mkurt__",x
    ," ,          (  o7__",x,"-  o5__",x,"+  o3__",x,"-  o1__",x,")/(  q3__",x,"-  q1__",x,")-1.23309511548521 AS   mkurt__",x
    ," ,GREATEST("
               ," (u_o7__",x,"-  o5__",x,"+  o3__",x,"-  o1__",x,")/(  q3__",x,"-  q1__",x,")"
               ,",(u_o7__",x,"-u_o5__",x,"+u_o3__",x,"-  o1__",x,")/(u_q3__",x,"-u_q1__",x,")"
               ,",(  o7__",x,"-  o5__",x,"+u_o3__",x,"-  o1__",x,")/(  q3__",x,"-u_q1__",x,")"
               ,",(  o7__",x,"-  o5__",x,"+  o3__",x,"-l_o1__",x,")/(  q3__",x,"-  q1__",x,")"
               ,",(  o7__",x,"-l_o5__",x,"+l_o3__",x,"-l_o1__",x,")/(l_q3__",x,"-l_q1__",x,")"
               ,",(  o7__",x,"-l_o5__",x,"+  o3__",x,"-  o1__",x,")/(l_q3__",x,"-  q1__",x,")"
                                                                                          ,")-1.23309511548521 AS u_mkurt__",x

# cskurt = (p97_5 - p2_5)/qrange - 2.90584695167016 : (normalized to 0 for a normal distribution, see Crow-Siddiqui, 1967)

    ," ,LEAST   ("
               ," (l_p97_5__",x,"-  p2_5__",x,")/(  q3__",x,"-  q1__",x,")"
               ,",(  p97_5__",x,"-  p2_5__",x,")/(  q3__",x,"-l_q1__",x,")"
               ,",(l_p97_5__",x,"-  p2_5__",x,")/(l_q3__",x,"-l_q1__",x,")"
               ,",(  p97_5__",x,"-u_p2_5__",x,")/(  q3__",x,"-  q1__",x,")"
               ,",(  p97_5__",x,"-  p2_5__",x,")/(u_q3__",x,"-  q1__",x,")"
               ,",(  p97_5__",x,"-u_p2_5__",x,")/(u_q3__",x,"-u_q1__",x,")"
                                                                       ,")-2.90584695167016 AS l_cskurt__",x
    ," ,          (  p97_5__",x,"-  p2_5__",x,")/(  q3__",x,"-  q1__",x,")-2.90584695167016 AS   cskurt__",x
    ," ,GREATEST("
               ," (u_p97_5__",x,"-  p2_5__",x,")/(  q3__",x,"-  q1__",x,")"
               ,",(  p97_5__",x,"-  p2_5__",x,")/(  q3__",x,"-u_q1__",x,")"
               ,",(u_p97_5__",x,"-  p2_5__",x,")/(u_q3__",x,"-u_q1__",x,")"
               ,",(  p97_5__",x,"-l_p2_5__",x,")/(  q3__",x,"-  q1__",x,")"
               ,",(  p97_5__",x,"-  p2_5__",x,")/(l_q3__",x,"-  q1__",x,")"
               ,",(  p97_5__",x,"-l_p2_5__",x,")/(l_q3__",x,"-l_q1__",x,")"
                                                                       ,")-2.90584695167016 AS u_cskurt__",x

#####################################
    ))

    dbSendQuery(.g$mysql.connection, "DROP TABLE IF EXISTS _true_and_extreme_results" ) 
    dbSendQuery(.g$mysql.connection, cpaste(
    
         " CREATE TABLE _true_and_extreme_results "
        ," AS SELECT t.*, ",stats1.extreme.variants.sql,", ",quantiles.sql,", ",stat.formulas3
        ," FROM _trueresults1 AS t, _quantiles_and_variants AS q, _stats1_extreme_variants AS v "
        ," WHERE "
            ,paste(
                "t.",g,"=q.",g," AND t.",g,"=v.",g
                ,sep="",collapse=" AND ")
        ," GROUP BY ",cpaste("t.",g)
    ))

    if (.g$.jk.debug) {
        print("_true_and_extreme_results")
        print(dbReadTable(.g$mysql.connection,"_true_and_extreme_results"))
    }

    .g$true.and.extreme.results <- "_true_and_extreme_results"
    .g$stat.names <- c(stat.names1,stat.names2,stat.names3)
    .g$stat.labels <- c(stat.labels1,stat.labels2,stat.labels3)
    
	.g <- .jk.univariate.finish(.g)
	
    if (return.format=="df") return(.g$return.df) else return(.g$return.mudit)
}
