#' Condstruct protection distributions based on specified quartiles and prepare a view with replacement values.
#' @param .g list of temporary objects needed.
#' @return Updated version of .g.
.jk.prepare.replacements <- function(.g){
	# requires:
	#	analyse
	#	mysql.connection
	# assigns:
	#	analyse.names
	#   analyse.labels
	#	replacements.sql1
	#	replacements.sql2

	# each entry in analyse is of the form
	#   list(name=col.name, quartiles=c(q0,q1,q2,q3,q4))
	# where q0,...,q4 are the "protection" quartiles 
	# from which the replacement distribution is constructed,
	# where q0 can be -Inf and q4 can be Inf


	# extract names and labels:
	
	.g$analyse.names <- 
		x <- unlist(lapply(.g$analyse,function(ai){ ai$name }))
	
	.g$analyse.labels <- 
		x <- unlist(lapply(.g$analyse,function(ai){ ai$label }))

	# the 3rd quartile of the standard normal distribution:
	qn <- qnorm(3/4)

		
	# determine replacement transformations y = f(x) with normally distributed x:

	sql1 <- list()
	sql2 <- list()
	
	for (i in 1:length(.g$analyse)) { # for each analysis variable:
	
		ai <- .g$analyse[[i]]
		name <- ai$name
		q0 <- ai$quartiles[[1]]
		q1 <- ai$quartiles[[2]]
		q2 <- ai$quartiles[[3]]
		q3 <- ai$quartiles[[4]]
		q4 <- ai$quartiles[[5]]
		
		if (q0==-Inf && q4==Inf) {  
		
			# unbounded: y = a + bx + c*sqrt(1+x²)
			b <- (q3-q1)/2 / qn
			c <- ((q3+q1)/2-q2) / (sqrt(1+qn^2)-1)
			a <- (q3+q1)/2 - c*sqrt(1+qn^2)
			sql1[[i]] <- cpaste( a,"+",b,"*n1__",name,"+",c,"*SQRT(1+POW(n1__",name,",2))" )
			sql2[[i]] <- cpaste( a,"+",b,"*n2__",name,"+",c,"*SQRT(1+POW(n2__",name,",2))" )
			
		} else if (q0>-Inf && q4==Inf) {  
		
			# only bounded from below: log-normal y = q0 + exp(a+bx), ignore q1
			a <- log(q2-q0)
			b <- log((q3-q0)/(q2-q0)) / qn
			sql1[[i]] <- cpaste( q0,"+EXP(",a,"+",b,"*n1__",name,")" )
			sql2[[i]] <- cpaste( q0,"+EXP(",a,"+",b,"*n2__",name,")" )
			
		} else if (q0==-Inf && q4<Inf) {  
		
			# only bounded from above: log-normal y = q4 - exp(a+bx), ignore q3
			a <- log(q4-q2)
			b <- log((q4-q1)/(q4-q2)) / qn
			sql1[[i]] <- cpaste( q4,"-EXP(",a,"+",b,"*n1__",name,")" )
			sql2[[i]] <- cpaste( q4,"-EXP(",a,"+",b,"*n2__",name,")" )	
								
		} else { 
		
			# bounded from both sides: y = a + b*atan(c+dx)
			if (q2-q0 <= q4-q2) {  
				# q2 nearer q0: ignore q1
				a <- (q4+q0)/2
				b <- (q4-q0) / pi
				c <- tan((q2-a)/b)
				d <- (tan((q3-a)/b) - c) / qn
			} else {  
				# q2 nearer q4: ignore q3
				a <- (q4+q0)/2
				b <- (q4-q0) / pi
				c <- tan((q2-a)/b)
				d <- (tan((q1-a)/b) - c) / (-qn)
			}
			sql1[[i]] <- cpaste( a,"+",b,"*ATAN(",c,"+",d,"*n1__",name,")" )
			sql2[[i]] <- cpaste( a,"+",b,"*ATAN(",c,"+",d,"*n2__",name,")" )
			
		}
		# Future work: add option for further restrictions such as rounding with offsets
	}
	
	# assign global variables:
	.g$replacements.sql1 <-
		sql1
	.g$replacements.sql2 <-
		sql2
		
	if (.g$.jk.debug) {
		print("replacements.sql1")
		print(.g$replacements.sql1)
		print("replacements.sql2")
		print(.g$replacements.sql2)
	}
    return(.g)
}
